open Xlist
open Xstring
open Xcoq
open Debug

open Interfaces

open Syntax_module_template

open Definitions_lift_subst
open Definitions_others

open Properties_lift_subst
open Properties_others

(** class of modules, ordered list of nodes *)
class gmodule_impl _name _coq_predefs _nodes _coq_postdefs : gmodule =
object (self)
	(** name of the module *)
	val name : string = _name
	(** verbatim preliminary definitions in coq *)
	val coq_predefs : string list = _coq_predefs
	(** list of elements *)
	val nodes : node list = _nodes
	(** verbatim appendix definitions in coq *)
	val coq_postdefs : string list = _coq_postdefs
	
	(** initializer : give reference to this gmodule to every sub-element *)
	initializer
	begin
		Debug.println "### gmodule initializer started.";
		List.iter (fun (n : node) -> n#set_gmodule (self :> gmodule)) nodes;
		Debug.println "### gmodule self linking done.";
		self#throughs_init;
		Debug.println "### grammars throughs initialization done.";
		self#throughs_expand;
		Debug.println "### grammars throughs expansion done.";
		Debug.print (self#to_string())
	end
	
	(** returns all the grammars of the gmodule *)
	method get_all_grams : grammar list =
		List.concat
			(List.map (fun n -> n#get_grams) nodes)
	
	(** returns all the grammars that has index in the gmodule *)
	method get_indexed_grams : grammar list =
		List.filter (fun g -> g#has_index) self#get_all_grams
	
	(** returns all the grammars that has index in the gmodule and
	  * that can be reached from the grammar of the given name *)
	method get_indexed_grams_through (gr : grammar) : grammar list =
		List.filter
			(fun g -> g = gr || List.mem g gr#get_throughs)
			self#get_indexed_grams
	
	(** returns the grammar of the given name as an option *)
	method get_gram (gname : string) : grammar =
		let rec aux l =
			match l with
					[] -> raise Not_found
				| g :: gs ->
						if g#get_name = gname then g else aux gs
		in
		aux self#get_all_grams
	
	(** returns true/false whether the grammar is one of the gmodule *)
	method is_gmodule_gram (gname : string) : bool =
		try (let _ = self#get_gram gname in true)
		with Not_found -> false
	
	(** initializes the through list, collecting immediate
	* subcategories that has index for every grammar of this module *)
	method private throughs_init : unit =
		List.iter (fun n -> n#throughs_init) nodes
	
	(** expands the through list, adding immediate
	* subcategories throughs *)
	method private throughs_expand : unit =
		(* to avoid infinite loop when debugging, we maximize the number of recursive calls *)
		let max_expand_count = 100 in
		let rec aux nb =
			if nb = max_expand_count then
				begin
					Debug.println "Ast.gmodule#throughs_expand : max_expand_count reached, exiting.";
					exit 1
				end else
				let result =
					List.fold_right
						(fun n -> fun result -> n#throughs_expand || result)
						nodes false
				in if result then aux (nb + 1) else ()
		in aux 0
	
	(** syntax definition generation *)
	method private syntax_gen (def : terms_definition) : string list =
		def#coq_predefs @ [""] @
		("  (** *** " ^ def#title ^ " *)") :: [""] @
		(Xlist.concat_with
				[""]
				(List.map (fun n -> n#syntax_gen def) nodes))
	
	(** syntax definition generation *)
	method private function_gen (def : fun_definition) : string list =
		let one_source_gen g =
			let _ =
				match g with
						Some gram -> def#set_source gram
					| None -> ()
			in
			def#coq_predefs @ [""] @
			("  (** *** " ^ def#title ^ " *)") :: [""] @
			(Xlist.concat_with
					[""]
					(List.map (fun n -> n#function_gen def) nodes))
		in
		List.concat
			(List.map one_source_gen def#get_source_grams)
	
	(** terms and functions definitions for a given template list *)
	method private terms_and_funs_def : string list =
		let rec aux_terms l =
			match l with
					[] -> []
				| def :: l -> (self#syntax_gen def) @ [""] @ aux_terms l
		and aux_fun l =
			match l with
					[] -> []
				| def :: l -> (self#function_gen def) @ [""] @ aux_fun l
		in
		(* primary terms, lift and subst definitions *)
		self#syntax_gen (new de_bruijn_terms (self :> gmodule)) @ [""] @
		self#function_gen ((new lift_fun (self :> gmodule)) :> fun_definition) @ [""] @
		self#function_gen ((new subst_fun (self :> gmodule)) :> fun_definition) @ [""] @
		(* additional definitions *)
		aux_terms (aux_terms_definitions (self :> gmodule)) @ [""] @
		aux_fun (aux_fun_definitions (self :> gmodule))
	
	(** Lifting and substitution primary properties (index properties) *)
	method private primary_lift_and_subst_props : string list =
		let to_define = self#get_indexed_grams in (* TODO : deplacer tout dans les proprietes. Supprimer to_define ?? *)
		
		let lift = (new lift_fun (self :> gmodule)) in
		
		let lift_index_lemmas_and_tacs =
			List.split
				(List.map
						(fun g -> lift#index_lemmas_and_tactics g) to_define) in
		let lift_index_lemmas = List.concat (fst lift_index_lemmas_and_tacs) in
		let lift_tactic_body = List.concat (snd lift_index_lemmas_and_tacs) in
		
		let lift_index_equation_tacs =
			List.split
				(List.map
						(fun g -> lift#index_equation_lemmas g) to_define) in
		let lift_index_equation_lemmas = List.concat (fst lift_index_equation_tacs) in
		let lift_equation_tactic_body = List.concat (snd lift_index_equation_tacs) in
		
		let subst = (new subst_fun (self :> gmodule)) in
		let subst_index_lemmas_and_tacs =
			List.split
				(List.map
						(fun g -> subst#index_lemmas_and_tactics g) to_define) in
		let subst_index_lemmas = List.concat (fst subst_index_lemmas_and_tacs) in
		let subst_tactic_body = List.concat (snd subst_index_lemmas_and_tacs) in
		
		let subst_index_equation_tacs =
			List.split
				(List.map
						(fun g -> subst#index_equation_lemmas g) to_define) in
		let subst_index_equation_lemmas = List.concat (fst subst_index_equation_tacs) in
		let subst_equation_tactic_body = List.concat (snd subst_index_equation_tacs) in
		
		let substs =
			("#lift_index_lemmas#", lift_index_lemmas) ::
			("#subst_index_lemmas#", subst_index_lemmas) ::
			("#lift_tactic_body#", lift_tactic_body) ::
			("#subst_tactic_body#", subst_tactic_body) ::
			
			("#lift_index_equation_lemmas#", lift_index_equation_lemmas) ::
			("#subst_index_equation_lemmas#", subst_index_equation_lemmas) ::
			("#lift_equation_tactic_body#", lift_equation_tactic_body) ::
			("#subst_equation_tactic_body#", subst_equation_tactic_body) ::
			[]
		and text =
			"#lift_index_lemmas#" ::
			"" ::
			"  Ltac lift_index_dec_tac :=" ::
			"    match goal with" ::
			"#lift_tactic_body#" ::
			"    end." ::
			"" ::
			"#subst_index_lemmas#" ::
			"" ::
			"  Ltac subst_index_dec_tac :=" ::
			"    match goal with" ::
			"#subst_tactic_body#" ::
			"    end." ::
			"" ::
			"#lift_index_equation_lemmas#" ::
			"" ::
			"  Ltac lift_index_tac :=" ::
			"    match goal with" ::
			"#lift_equation_tactic_body#" ::
			"    end." ::
			"" ::
			"#subst_index_equation_lemmas#" ::
			"" ::
			"  Ltac subst_index_tac :=" ::
			"    match goal with" ::
			"#subst_equation_tactic_body#" ::
			"    end." ::
			"" ::
			"  Ltac #index_tac# :=" ::
			"    repeat (progress (lift_index_tac || subst_index_tac))." ::
			[] in
		Xstring.replaces [] substs text
	
	(** property generation *)
	method property_gen (prop : property)
	: string list * (string * (string list * string list)) list =
		let (lemmas, tactics) =
			List.split
				(List.map (fun n -> n#property_gen prop) nodes) in
		(Xlist.concat_with [""] lemmas, List.concat tactics)
	
	(** properties generation for all chapters *)
	method private all_properties_gen : string list =
		let rec aux l =
			match l with
					[] -> []
				| chap :: l -> chap#generate @ [""] @ aux l in
		(* main lift and subst properties *)
		(new lift_chapter (self :> gmodule))#generate @ [""] @
		(new subst_chapter (self :> gmodule))#generate @ [""] @
		(* additional properties *)
		aux (aux_properties (self :> gmodule))
	
	(** coq output generation *)
	method coq_output : string list =
		(** Global names used in all definitions and properties *)
		let subst =
			("#gmodule#", name) ::
			("#database#", name ^ "_database") ::
			("#crush#", "crush_tac") ::
			("#ecrush#", "ecrush_tac") ::
			("#index_tac#", "index_tac") ::
			[] in
		
		(* Coq module header and footer *)
		let header = SyntaxModuleTemplate.header
		and footer = SyntaxModuleTemplate.footer in
		let _ = Debug.println "### Module header and footer invoked." in
		
		let title_def title def =
			Xcoq.coq_title 1 title @ [""] @ def @ [""] in
		
		let terms_and_funs_def = self#terms_and_funs_def in
		
		let primary_lift_and_subst_props = self#primary_lift_and_subst_props in
		
		let all_properties = self#all_properties_gen in
		
		let output =
			(* header and user pre-definitions *)
			header @ [""] @
			coq_predefs @ [""] @
			
			(title_def "Inductive syntax and function definitions"
					terms_and_funs_def) @
			
			(title_def "Primary lifting and substitution properties"
					primary_lift_and_subst_props) @
			
			(title_def "Additional properties"
					all_properties) @
			
			(title_def "Main dbgen tactic"
					("  Ltac dbgen_tac :=" ::
					 "    #crush#;" ::
					 "    repeat (progress (#index_tac# || lift_tac || subst_tac || fv_permut_tac))." ::
					 [])) @
			
			(* user post-definitions and footer *)
			coq_postdefs @ [""] @
			footer in
		Xstring.replaces subst [] output
	
	(** Pretty printing *)
	method to_string () : string =
		"Module " ^ name ^ ".\n\n" ^
		(String.concat "\n" coq_predefs) ^ "\n" ^
		(String.concat
				"\n"
				(List.map (fun n -> n#to_string ()) nodes)) ^
		(String.concat "\n" coq_postdefs) ^ "\n" ^
		"\nEnd " ^ name ^ ".\n"
end
